import random
from Schedule import *

class Tester(Schedule):
    """
    Tester(
        file_name,
        sorting function,
        sequential/parallel function,
        solver_selector)

    sequential/parallel function(job) is a function that takes a
    job and returns a boolean. If it returns true then the job is run
    in parallel with all the solvers (random first).
    Otherwise the job is run sequentially using the fs first.

    solver_selector:
        'random' -> random selection
        'oracle' -> fast solver first
        'fs' -> first the solver selecte by fs classifier, than random order
        'satzilla' -> order the solver according to the expected execution times
    """

    def __init__(self,*args):
        super(Tester,self).__init__(args[0],args[1])
        self.parallel_function = args[2]
        self.solver_selector = args[3]
            
    def create_jobs(self):
        self.current_job = 0
        #jobs that are ready to be sent
        self.jobs_to_run = {}
        #jobs that are not sumbitted
        self.jobs_running = {}
        #jobs that are running
        self.other_solvers = {}
        #jobs -> solvers to use in case the fs fails
        self.jobs_to_rerun = []
        #jobs that should be rerun because the fs failed

        if self.solver_selector == 'fs':
            for i in range(len(self.instances)):
                solvers = ["mistral", "choco", "abscon"]
                solvers.remove(self.instances[i].fs)
                random.shuffle(solvers)
                if self.parallel_function(self.instances[i]):
                    self.jobs_to_run[i] = solvers
                    self.jobs_to_run[i].append(self.instances[i].fs)
                else:
                    self.other_solvers[i] = solvers
                    self.jobs_to_run[i] = [self.instances[i].fs]
                self.jobs_running[i] = []
        elif self.solver_selector == 'oracle':
            for i in range(len(self.instances)):
                solvers = ["mistral", "choco", "abscon"]
                solvers.remove(self.instances[i].get_best_solver())
                random.shuffle(solvers)
                if self.parallel_function(self.instances[i]):
                    self.jobs_to_run[i] = solvers
                    self.jobs_to_run[i].append(self.instances[i].get_best_solver())
                else:
                    self.other_solvers[i] = solvers
                    self.jobs_to_run[i] = [self.instances[i].get_best_solver()]
                self.jobs_running[i] = []
        elif self.solver_selector == 'random':
            for i in range(len(self.instances)):
                solvers = ["mistral", "choco", "abscon"]
                random.shuffle(solvers)
                if self.parallel_function(self.instances[i]):
                    self.jobs_to_run[i] = solvers
                else:
                    self.jobs_to_run[i] = [solvers.pop()]
                    self.other_solvers[i] = solvers
                self.jobs_running[i] = []
        elif self.solver_selector == 'satzilla':
            for i in range(len(self.instances)):
                solvers = self.instances[i].get_satzilla_solver_order()
                solvers.reverse()
                if self.parallel_function(self.instances[i]):
                    self.jobs_to_run[i] = solvers
                else:
                    self.jobs_to_run[i] = [solvers.pop()]
                    self.other_solvers[i] = solvers
                self.jobs_running[i] = []
        else:
            print "Error in solver_selector parameter"
            print "Chosen " + self.solver_selector
            print "Expected random, oracle, satzilla or fs"
                    

    def getjob(self,client):
        if len(self.jobs_to_rerun) != 0:
            #job that should rerun (fs failed)
            job = self.jobs_to_rerun.pop(0)
            self.jobs_running[job].append(client)
            return (job,self.other_solvers[job].pop(),TIMEOUT)

        if self.current_job >= len(self.instances):
            return (NO_JOB,"",TIMEOUT)
        else:
            job = self.current_job
            if job in self.solving_times: #job already solved
                self.current_job += 1
                return self.getjob(client)
            else:
                solver = self.jobs_to_run[job].pop()
                if len(self.jobs_to_run[job]) == 0:
                    del self.jobs_to_run[job]
                    self.current_job += 1
                self.jobs_running[job].append(client)
                return (job,solver,TIMEOUT)

    def finished(self,client,job):
        #print "%s: finished job %s by client %s" % (now(),job,client.name)
        if job in self.solving_times:
            #this should never happen in the simulation
            print "Error: job %s has not been interrupted (client %s)" % (job,client.name)
        else:
            super(Tester,self).finished(client, job)
            self.jobs_running[job].remove(client)
            i = Interruptor()
            i.stop(self.jobs_running[job])
            self.jobs_running[job] = []
            

    def not_finished(self,client,job):
        #print "%s: not finished job %s by client %s" % (now(), job,client.name)
        super(Tester,self).not_finished(client, job)
        self.jobs_running[job].remove(client)
        if ((job in self.other_solvers) and (len(self.other_solvers[job]) > 0)):
            self.jobs_to_rerun.append(job)