from SimPy.Simulation import *
from Schedule import *
from itertools import *
from elaborate_files import *
from Constants import *
import string
import time


from Tester import *
from TesterMistral import *
from TesterCpHydra import *
from TesterSatzilla import *

import scipy.stats as sp
import numpy as np

import csv
import os


################################################################
# Schedules to test
################################################################

"""
Collection of auxiliary functions used to concatenate sorting
functions or to create additional test
"""

def concatenate(lists):
        return reduce( list.__add__, lists)

def create_orders_aux(schedule,schedule_name_pre,schedule_name_post,funct,funct_names, other_parameter):
    def aux(pairs):
        functs, names = zip(*pairs)
        return [
            schedule_name_pre + string.join(names, "<") + schedule_name_post,
            schedule,
            [compose_sort(list(functs))] + other_parameter
            ]
            
    per = permutations(zip(funct,funct_names))
    return map(aux,per)

def create_orders(schedule,schedule_name_pre,schedule_name_post,funct,funct_names, other_parameter):
    def allsubsets(ls):
        return concatenate(map(
            lambda x: list(combinations(ls,x)),
            range(1,len(ls)+1)))

    return concatenate(map(
        lambda x : create_orders_aux(schedule,schedule_name_pre,schedule_name_post,x[0],x[1], other_parameter),
        zip(allsubsets(funct), allsubsets(funct_names))
        ))

"""
baselines is the list of the tests. It contains a tuple containing
the following parameters:
    1) name of the test
    2) object to call for the test (Tester, MistralTester or CpHydraTester)
    3) parameters to use for the object performing the test (see classes)
"""

best_instance_selector = compose_sort([stdv_key, threec_key, avg_key, max_key])

baselines = []

baselines.append([
    "oracle_oracle",
    Tester,
    [compose_sort([fast_key]),
    lambda x: False,
    'oracle'
    ]])

baselines.append([
    "oracle_fs",
    Tester,
    [compose_sort([fast_key]),
    lambda x: False,
    'fs'
    ]])

baselines.append([
    "oracle_random",
    Tester,
    [compose_sort([fast_key]),
    lambda x: False,
    'random'
    ]])

baselines.append([
    "oracle_satzilla+",
    Tester,
    [compose_sort([fast_key]),
    lambda x: False,
    'satzilla'
    ]])

baselines.append([
    "oracle_mistral",
    TesterMistral,
    [compose_sort([fast_key])]
    ])

baselines.append([
    "oracle_cphydra",
    TesterCpHydra,
    [compose_sort([fast_key])]
    ])

baselines.append([
    "oracle_satzilla",
    TesterSatzilla,
    [compose_sort([fast_key])]
    ])

baselines.append([
    "random_oracle",
    Tester,
    [random.shuffle,
    lambda x: False,
    'oracle'
    ]])

baselines.append([
    "random_fs",
    Tester,
    [random.shuffle,
    lambda x: False,
    'fs'
    ]])

baselines.append([
    "random_random",
    Tester,
    [random.shuffle,
    lambda x: False,
    'random'
    ]])

baselines.append([
    "random_satzilla+",
    Tester,
    [random.shuffle,
    lambda x: False,
    'satzilla'
    ]])

baselines.append([
    "random_mistral",
    TesterMistral,
    [random.shuffle]
    ])

baselines.append([
    "random_cphydra",
    TesterCpHydra,
    [random.shuffle]
    ])

baselines.append([
    "random_satzilla",
    TesterSatzilla,
    [random.shuffle]
    ])

baselines.append([
    "satzilla_oracle",
    Tester,
    [compose_sort([regft_key]),
    lambda x: False,
    'oracle'
    ]])

baselines.append([
    "satzilla_fs",
    Tester,
    [compose_sort([regft_key]),
    lambda x: False,
    'fs'
    ]])

baselines.append([
    "satzilla_random",
    Tester,
    [compose_sort([regft_key]),
    lambda x: False,
    'random'
    ]])

baselines.append([
    "satzilla_satzilla+",
    Tester,
    [compose_sort([regft_key]),
    lambda x: False,
    'satzilla'
    ]])

baselines.append([
    "satzilla_satzilla",
    TesterSatzilla,
    [compose_sort([regft_key])]
    ])

baselines.append([
    "satzilla_cpHydra",
    TesterCpHydra,
    [compose_sort([regft_key])]
    ])

baselines.append([
    "satzilla_Mistral",
    TesterMistral,
    [compose_sort([regft_key])]
    ])

baselines.append([
    "I_oracle",
    Tester,
    [best_instance_selector,
    lambda x: False,
    'oracle'
    ]])

baselines.append([
    "I_random",
    Tester,
    [best_instance_selector,
    lambda x: False,
    'random'
    ]])

baselines.append([
    "I_satzilla+",
    Tester,
    [best_instance_selector,
    lambda x: False,
    'satzilla'
    ]])

baselines.append([
    "I_fs",
    Tester,
    [best_instance_selector,
    lambda x: False,
    'fs'
    ]])
    
baselines.append([
    "I_mistral",
    TesterMistral,
    [best_instance_selector]
    ])

baselines.append([
    "I_cphydra",
    TesterCpHydra,
    [best_instance_selector]
    ])

baselines.append([
    "I_satzilla",
    TesterSatzilla,
    [best_instance_selector]
    ])

"""
baselines.append([
    "mistralC_mistral",
    TesterMistral,
    [compose_sort([mistral_key])]
    ])

baselines.append([
    "mistralC_oracle",
    Tester,
    [compose_sort([mistral_key]),
    lambda x: False,
    'oracle'
    ]])

baselines.append([
    "mistralC_random",
    Tester,
    [compose_sort([mistral_key]),
    lambda x: False,
    'random'
    ]])

baselines.append([
    "mistralC_fs",
    Tester,
    [compose_sort([mistral_key]),
    lambda x: False,
    'fs'
    ]])
    
"""   

"""parallel"""

"""
baselines.append([
    "Strategy_FS_selection1",
    Interrupt1,
    [compose_sort([threec_key, max_key, avg_key, stdv_key]),
    selection_1]
    ])

baselines.append([
    "Strategy_FS_selection2",
    Interrupt1,
    [compose_sort([threec_key, max_key, avg_key, stdv_key]),
    selection_2]
    ])

baselines.append([
    "Strategy_FS_selection3",
    Interrupt1,
    [compose_sort([threec_key, max_key, avg_key, stdv_key]),
    selection_3]
    ])

baselines.append([
    "Strategy_FS_all_parallel",
    Interrupt1,
    [compose_sort([threec_key, max_key, avg_key, stdv_key]),
    lambda x: True]
    ])

baselines.append([
    "SJF_FS_all_parallel",
    Interrupt1,
    [compose_sort([fast_key]),
    lambda x: True]
    ])

baselines.append([
    "SJF_FS_selection2",
    Interrupt1,
    [compose_sort([fast_key]),
    selection_2]
    ])

baselines.append([
    "random_FS_selection2",
    Interrupt1,
    [random.shuffle,
    selection_2]
    ])
"""

"""
simulation is the list of the test to perform.
It is a list of tuples like baseline
"""

simulations = []


simulations.append([
    "3C<avg<stdv<max_fs",
    Tester,
    [compose_sort([threec_key, max_key, avg_key, stdv_key]),
    lambda x: False,
    'fs'
    ]])

"""
simulations.extend(
    create_orders(
        Tester,
        "",
        "_fs",
        [threec_key, avg_key, stdv_key, max_key],
        ["3C", "avg", "stdv", "max"],
        [lambda x: False, 'fs']
        )
    )

simulations.extend(
    create_orders(
        Tester,
        "",
        "_fs",
        [invert(threec_key), invert(avg_key), stdv_key, invert(max_key)],
        ["{\\bar 3C}", "{\\bar avg}", "stdv", "{\\bar max}"],
        [lambda x: False, 'fs']
        )
    )

simulations.extend(
    create_orders(
        Tester,
        "",
        "_fs",
        [invert(threec_key), invert(avg_key), invert(stdv_key), invert(max_key)],
        ["{\\bar 3C}", "{\\bar avg}", "{\\bar stdv}", "{\\bar max}"],
        [lambda x: False, 'fs']
        )
    )

"""


simulations.sort(key= lambda x: x[0])
last = simulations[-1]
for i in range(len(simulations)-2, -1, -1):
    if last[0] == simulations[i][0]:
        del simulations[i]
    else:
        last = simulations[i]

################################################################
# Functions definitions
################################################################

"""
simulate_fold perform the test on a single fold
"""

def simulate_fold(file_name, output_file, (id, schedule, args)):
    initialize()
    s = schedule(file_name,*args)
    s.create_jobs()
    for i in range(NCLIENTS):
        c = Client(str(i),s)
        activate(c,c.start())
    simulate(until=10000000.0)

    jobs = len(s.instances)
    makespan = max([i for i in s.finishing_times.values()])
    sum_running_times = sum([i for i in s.running_times.values()])
    sum_solving_times = sum([i for i in s.solving_times.values()])
    output_file.writerow([
    sum_solving_times/jobs,
    makespan,
    np.median(s.solving_times.values()),
    sum_running_times/NCLIENTS,
    (makespan-sum_running_times/NCLIENTS)/makespan,
    len(s.solving_times),
    jobs,
    s.interrupts_num,
    s.solver_calls["mistral"],
    s.solver_calls["choco"],
    s.solver_calls["abscon"],
    ])

"""
write_first_line prints the first line of the csv
containing the results of the tests
"""
def write_first_line(file):
    file.writerow([
        "average solving time",
        "makespan",
        "median solving time"
        "average processor use",
        "% unused",
        "instances solved",
        "total instances",
        "interrupts number",
        "mistralCalls",
        "chocoCalls",
        "absconCalls"])

"""
process_files process the file obtained using RapidMiner
adding some information like the CpHydra solving time.
It may require a long to compute and it is not necessary to
call it again after the firs process has been done.
"""

def process_files():
    files = os.listdir(TEST_FILES_PATH)
    files.sort()
    for f in files:
        print "processing file " + f
        elaborate_files(f)

"""
main function that computes all the test.
Needs to be called when all the files obtained using RapidMiner
have been processed via the function call process_files
"""
def run_simulations():
    files = filter( lambda x: x.endswith("_all"),os.listdir(PRED_FILES_PATH))
    files.sort()

    global out_files_names
    out_files_names = {}
    global out_files_pointers
    out_files_pointers = {}
    cvs_writer = {}

    for i in simulations:
        out_files_names[i[0]] = LOGS_PATH + i[0] + ".csv"
        out_files_pointers[i[0]] = open(out_files_names[i[0]], 'wb')
        cvs_writer[i[0]] = csv.writer(out_files_pointers[i[0]])
        write_first_line(cvs_writer[i[0]])

    for i in baselines:
        out_files_names[i[0]] = LOGS_PATH + i[0] + ".csv"
        out_files_pointers[i[0]] = open(out_files_names[i[0]], 'wb')
        cvs_writer[i[0]] = csv.writer(out_files_pointers[i[0]])
        write_first_line(cvs_writer[i[0]])
        
    for i in files:
        print "Simulation of %s" %(i)
        for j in simulations:
            #print "todel simultate %s" %(j[0])
            simulate_fold(PRED_FILES_PATH + i, cvs_writer[j[0]], j)
        for j in baselines:
            #print "todel simultate %s" %(j[0])
            simulate_fold(PRED_FILES_PATH + i, cvs_writer[j[0]], j)
        
    print "Simulation terminated"

    for i in simulations:
        out_files_pointers[i[0]].close()

    for i in baselines:
        out_files_pointers[i[0]].close()


"""
elaborate the data to get some statistics like the average solving time
"""
def compute_statistics():

    global averages
    global medians
    global mkspan
    global mistral_calls
    global abscon_calls
    global choco_calls

    averages = {}
    medians = {}
    mkspan = {}
    mistral_calls = {}
    abscon_calls = {}
    choco_calls = {}

    for i in simulations:
        averages[i[0]] = []
        mkspan[i[0]] = []
        medians[i[0]] = []
        mistral_calls[i[0]] = 0
        abscon_calls[i[0]] = 0
        choco_calls[i[0]] = 0
        reader = csv.reader(open(out_files_names[i[0]], "rb"))
        row_number = 0
        for row in reader:
            if row_number != 0:
                averages[i[0]].append(float(row[0]))
                mkspan[i[0]].append(float(row[1]))
                medians[i[0]].append(float(row[2]))
                mistral_calls[i[0]] += int(row[-3])
                abscon_calls[i[0]] += int(row[-1])
                choco_calls[i[0]] += int(row[-2])
            else:
                row_number += 1

    for i in baselines:
        averages[i[0]] = []
        mkspan[i[0]] = []
        medians[i[0]] = []
        mistral_calls[i[0]] = 0
        abscon_calls[i[0]] = 0
        choco_calls[i[0]] = 0
        reader = csv.reader(open(out_files_names[i[0]], "rb"))
        row_number = 0
        for row in reader:
            if row_number != 0:
                averages[i[0]].append(float(row[0]))
                mkspan[i[0]].append(float(row[1]))
                medians[i[0]].append(float(row[2]))
                mistral_calls[i[0]] += int(row[-3])
                abscon_calls[i[0]] += int(row[-1])
                choco_calls[i[0]] += int(row[-2])
            else:
                row_number += 1

    global average
    global median_avg
    global std
    global avg_median
    global mkspan_average
    global mkspan_median

    average = {}
    median_avg = {}
    std = {}
    avg_median = {}
    mkspan_average = {}
    mkspan_median = {}

    for i in simulations:
        average[i[0]] = np.mean(averages[i[0]])
        median_avg[i[0]] = np.median(averages[i[0]])
        avg_median[i[0]] = np.mean(medians[i[0]])
        std[i[0]] = np.std(averages[i[0]])
        mkspan_average[i[0]] = np.mean(mkspan[i[0]])
        #mkspan_average[i[0]] = sum(mkspan[i[0]]) / len(mkspan[i[0]])
        mkspan_median[i[0]] = np.median(mkspan[i[0]])

    for i in baselines:
        average[i[0]] = np.mean(averages[i[0]])
        median_avg[i[0]] = np.median(averages[i[0]])
        avg_median[i[0]] = np.mean(medians[i[0]])
        std[i[0]] = np.std(averages[i[0]])
        mkspan_average[i[0]] = np.mean(mkspan[i[0]])
        #mkspan_average[i[0]] = sum(mkspan[i[0]]) / len(mkspan[i[0]])
        mkspan_median[i[0]] = np.median(mkspan[i[0]])


"""
print a table using a comma as separator summarizing the test results
"""
def print_full_table():

    best_simulation_avg = min(average.keys(), key=lambda x:average[x])
    best_simulation_mkspan = min(mkspan_average.keys(), key=lambda x:mkspan_average[x])

    print "best for minimizing averages: %s" % best_simulation_avg
    print "best for minimizing makespan: %s" % best_simulation_mkspan

    print "order , average, median, avgMedian, std, mkspanAverage, mkspanMedian, significanceBestAvg, significanceBestMkspan, mistral, choco, abscon"

    simulations.extend(baselines)
    for i in simulations:
        (t,p_1) = sp.ttest_rel(averages[i[0]], averages[best_simulation_avg])
        (t,p_2) = sp.ttest_rel(mkspan[i[0]], mkspan[best_simulation_mkspan])
        print "%s," % (i[0]),
        print "%s," % (average[i[0]]),
        print "%s," % (avg_median[i[0]]),
        print "%s," % (median_avg[i[0]]),
        print "%s," % (std[i[0]]),
        print "%s," % (mkspan_average[i[0]]),
        print "%s," % (mkspan_median[i[0]]),
        if p_1 < 0.05:
            print "1,",
        else:
            print "0,",
        if p_2 < 0.05:
            print "1,",
        else:
            print "0,",
        print "%s," % (mistral_calls[i[0]]),
        print "%s," % (choco_calls[i[0]]),
        print "%s" % (abscon_calls[i[0]])

"""
prints a latex table summarizing the results.
The base_test is the name of the test that is used to 
check the statistacal significance of the other tests
"""
def print_latex_table(test_list,base_test):

    list = [key for (value,key) in sorted([(average[key],key) for key in test_list])]

    print "\\begin{table*}"
    print "\\caption{caption}"
    print "\\centering"
    print "\\begin{tabular}{|cc|rl|}"
    print "\\hline"
    print "instance selector & solver selector & average & \\\\"
    print "\\hline"
    for i in list:
        (inst_selector, unused_ ,solver_selector) = i.rpartition('_')
        if "<" in inst_selector:
            print "$%s$ &" % inst_selector,
        else:
            print inst_selector + " &",
        print solver_selector + " &",
        mapping = {
            "satzilla": "\\satzillaNotSolved",
            "cphydra": "\\cphydraNotSolved",
            "mistral": "\\mistralNotSolved",
            }
        if  solver_selector in mapping:
            print "$>$ %.2f &" % average[i],
        else:
            print "%.2f &" % average[i],
        (t,p) = sp.ttest_rel(averages[i], averages[base_test])
        if (p < 0.05) & (average[i] <= average[base_test]):
            print "$\\circ$ ",
        elif (p < 0.05):
            print "$\\bullet$ ",
#        else:
#            print "& ",
#        if  solver_selector in mapping:
#            print mapping[solver_selector],
        print "\\\\"
    print "\\hline"
    print "\end{tabular}"
    print "\end{table*}"


    

#elaborate files obtained using rapidminer
process_files()
print "Start simulation"
globaltime = time.time()
run_simulations()
print "Time spend for tests: %s" % (time.time() - globaltime)
compute_statistics()
#print_full_table()

print_latex_table([
    "oracle_oracle",
    "I_oracle",
    "satzilla_oracle",
    "random_oracle",
    ],
    "I_oracle")

print ""
print_latex_table([
    "oracle_random",
    "I_random",
    "satzilla_random",
    "random_random",
    ],
    "I_random")

print ""
print_latex_table([
    "oracle_oracle",
    "oracle_fs",
    "oracle_satzilla",
    "oracle_satzilla+",
    "oracle_mistral",
    "oracle_cphydra",
    "oracle_random"
    ],
    "oracle_fs")

print ""
print_latex_table([
    "random_oracle",
    "random_fs",
    "random_satzilla",
    "random_satzilla+",
    "random_mistral",
    "random_cphydra",
    "random_random"
    ],
    "random_fs")

print ""
print_latex_table([
    "I_oracle",
    "I_fs",
    "I_satzilla",
    "I_satzilla+",
    "I_mistral",
    "I_cphydra",
    "I_random"
    ],
    "I_fs")

print ""
print_latex_table([
    "oracle_fs",
    "I_fs",
    "satzilla_fs",
    "random_fs",
    ],
    "I_fs")

print ""
print_latex_table([x[0] for x in simulations], "I_fs")
